from typing import get_type_hints, Annotated, get_origin, get_args, Generic, TypeVar

from .app import *
from .exceptions import *
from .module import *


T = TypeVar('T', bound='Parameter')


class Parameter(Generic[T]):
    def __call__(self) -> T:
        return self


class Provide(Option):
    def __init__(self, *args):
        self._targets = {}
        for arg in args:
            if isinstance(arg, Provider):
                hints = get_hints(arg.provider.__init__)
                self.init_provider(arg, hints, arg.provider)
                continue

            # check if it is a class of any type
            if isinstance(arg, type):
                hints = get_hints(arg.__init__)
                self._targets[arg] = ProvideTarget(
                    callable=arg, provides=arg, requires=get_requires_from_hints(
                        hints)
                )
                continue

            if callable(arg):
                hints = get_hints(arg)
                if "return" not in hints:
                    raise MissingHintError(
                        "Provide target must have a return type hint"
                    )

                if isinstance(hints["return"], Provider):
                    self.init_provider(hints["return"], hints, arg)
                    continue

                if isinstance(hints["return"], type):
                    self._targets[hints["return"]] = ProvideTarget(
                        callable=arg,
                        provides=hints["return"],
                        requires=get_requires_from_hints(hints),
                    )
                    continue

            raise PyDITypeError(
                "Provide target must be a callable constructor, a class or a Provider: ", arg)

    def init_provider(self, arg: Provider, hints: dict, call: callable):
        if not arg.group:
            self._targets[arg.name] = ProvideTarget(
                callable=call,
                provides=arg.name,
                requires=get_requires_from_hints(hints),
            )

        else:
            if arg.name not in self._targets:
                self._targets[arg.name] = []

            self._targets[arg.name].append(
                ProvideTarget(
                    callable=call,
                    provides=arg.name,
                    requires=get_requires_from_hints(hints),
                )
            )

    def apply(self, mod: Module):
        mod.add_provides(self._targets)


def get_hints(arg):
    # check if has annotations
    if issubclass(type(arg), Parameter):
        return {"return": type(arg)}
    if not hasattr(arg, "__annotations__"):
        return get_type_hints(arg)
    hints = arg.__annotations__
    for key, value in hints.items():
        if get_origin(value) == Annotated:
            for a in value.__metadata__:
                if isinstance(a, Annotations):
                    hints[key] = a.to_provider(get_args(value)[0])
                    break
    return hints


class Invoke(Option):
    def __init__(self, *args):
        self._targets = []
        for arg in args:
            if not callable(arg):
                raise PyDITypeError("Provide target must be callable")

            hints = get_hints(arg)

            self._targets.append(
                InvokeTarget(
                    callable=arg, requires=get_requires_from_hints(hints))
            )

    def apply(self, mod: Module):
        mod.add_invokes(self._targets)


def get_requires_from_hints(hints) -> List[str]:
    requires = []
    for key, value in hints.items():
        if key == "return":
            continue
        else:
            requires.append(value)
    return requires



class Hook:
    def __init__(self, on_start: callable = None, on_stop: callable = None):
        self.on_start = on_start
        self.on_stop = on_stop


class Lifecycle:
    def __init__(self) -> None:
        self.hooks = []

    def append_hook(self, hook: Hook):
        self.hooks.append(hook)

    def start(self):
        for hook in self.hooks:
            try:
                if callable(hook.on_start):
                    hook.on_start()
            finally:
                if callable(hook.on_stop):
                    hook.on_stop()



class PyDIException(Exception):
    pass


class PyDITypeError(PyDIException, TypeError):
    pass


class MissingHintError(PyDIException):
    pass


class MissingDependencyError(PyDIException):
    pass


class DependencyTypeError(PyDIException):
    pass



import inspect
from os.path import normpath
from typing import Any, Dict, List, Type

from .exceptions import *
from .lifecycle import *


class ProvideTarget:
    # This is a class to store the dependencies definitions within a module.
    def __init__(self, callable: callable, provides: str, requires: List[str]):
        self.callable = callable
        self.provides = provides
        self.requires = requires


ProviderKey = str | Type
ProvidesType = Dict[ProviderKey, ProvideTarget]
ConternerTargetType = object | List[object]
ContainerType = Dict[ProviderKey, ConternerTargetType]


class InvokeTarget:
    # This is a class to store the invoke definitions within a module.
    def __init__(self, callable: callable, requires: List[str]):
        self.callable = callable
        self.requires = requires


class Provider:
    # This is a class that is used to build advanced dependencies definitions.
    def __init__(self, provider: object, name: str | None = None, group: bool = False):
        self.provider = provider
        self.name = name
        if name is None:
            self.name = provider
        self.group = group


class Annotations:
    def __init__(self, name: str | None = None, group: bool = False):
        self.name = name
        self.group = group

    def to_provider(self, provider: object) -> Provider:
        return Provider(provider, name=self.name, group=self.group)


class Option:
    # This is the base class for the functional options pattern.
    def apply(self, mod: "Module"):
        pass


class Module(Option):
    # This is a class that is used to store all the dependencies and invoke targets,
    # and also holds the lifecycle object and the built dependencies.
    def __init__(self, *args):
        self.lifecycle = Lifecycle()
        self._provides: ProvidesType = {}
        self._invokes: List[InvokeTarget] = []

        for option in args:
            if not isinstance(option, Option):
                raise PyDITypeError(
                    "App constructor arguments must be of type Option")
            option.apply(self)

    def apply(self, mod: "Module"):
        mod.add_provides(self._provides)
        mod.add_invokes(self._invokes)

    def register_container(self, container: ContainerType) -> None:
        self.container = container

    def set_provides(self, provides: ProvidesType) -> None:
        self._provides = provides

    def get_provides(self) -> ProvidesType:
        return self._provides

    def build_dependencies(self, target: ProvideTarget | InvokeTarget) -> List:
        inject = []
        for require in target.requires:
            if require == Lifecycle:
                inject.append(self.lifecycle)
                continue

            key = require
            if isinstance(require, Provider):
                key = require.name

            provides_target = key_in(key, self._provides)
            if provides_target is None:
                raise MissingDependencyError(
                    f"Cannot find dependency: {key}, {self._provides}"
                )

            target = key_in(key, self.container)
            if target is not None:
                if isinstance(require, Provider) and inspect.isclass(target) and not issubclass(
                    target, require.provider
                ):
                    raise DependencyTypeError(
                        "Dependency is of wrong type",
                        key,
                        target,
                        "but should be of type",
                        require.provider,
                    )
                inject.append(target)
            else:
                provider = self.build_provider(provides_target)
                if isinstance(require, Provider):
                    if not require.group and not issubclass(
                        provider.__class__, require.provider
                    ):
                        raise DependencyTypeError(
                            "Dependency is of wrong type",
                            key,
                            provider.__class__,
                            "but should be of type",
                            require.provider,
                        )
                    if require.group and not issubclass(
                        provider[0].__class__, require.provider
                    ):
                        raise DependencyTypeError(
                            "Dependency is of wrong type",
                            key,
                            provider[0].__class__,
                            "but should be of type",
                            require.group,
                        )
                self.container[key] = provider
                inject.append(provider)

        return inject

    def build_provider(
        self, target: ProvideTarget | List[ProvideTarget]
    ) -> ConternerTargetType:
        if isinstance(target, ProvideTarget):
            inject = self.build_dependencies(target)
            try:
                return target.callable(*inject)
            except TypeError as e:
                if "missing" in str(e):
                    raise MissingHintError(
                        str(e), "verify if all hints are set")
                raise e

        elements = []
        for element in target:
            if not isinstance(element, ProvideTarget):
                raise PyDITypeError(
                    "Provide target list element must be of type ProvideTarget"
                )
            inject = self.build_dependencies(element)
            try:
                elements.append(element.callable(*inject))
            except TypeError as e:
                if "missing" in str(e):
                    raise MissingHintError(
                        str(e), "verify if all hints are set")
                raise e
        return elements

    def add_provides(self, targets: ProvidesType):
        if not isinstance(targets, Dict):
            raise PyDITypeError("Provide targets must be a dict")
        for value in targets.values():
            if not isinstance(value, ProvideTarget) and not isinstance(value, List):
                raise PyDITypeError(
                    "Provide target must be of type ProvideTarget or list of ProvideTarget"
                )
            if isinstance(value, List):
                for target in value:
                    if not isinstance(target, ProvideTarget):
                        raise PyDITypeError(
                            "Provide target list element must be of type ProvideTarget"
                        )

        self._provides = self._provides | targets

    def add_invokes(self, targets: List[InvokeTarget]):
        for target in targets:
            if not isinstance(target, InvokeTarget):
                raise PyDITypeError(
                    "Invoke target must be of type InvokeTarget")
        self._invokes.extend(targets)


def key_in(key: ProviderKey, elements: Dict[ProviderKey, Any]) -> ProvideTarget:
    if isinstance(key, str):
        return elements.get(key, None)

    elif isinstance(key, Type):
        for k, v in elements.items():
            if not isinstance(k, Type):
                continue
            # checks if class key is the same as k
            if k.__name__ == key.__name__ and normpath(inspect.getfile(k)) == normpath(
                inspect.getfile(key)
            ):
                return v
    else:
        raise PyDITypeError("Key must be of type str or a class")
    return None



import unittest
from typing import Annotated

from src.dinjections import *


class TestClass1:
    pass


class TestClass2:
    pass


class TestClass3:
    pass

    def run(self):
        pass


def new_test_class_3(
    p: Annotated[TestClass1, Annotations("t1")], p2: Annotated[TestClass1, Annotations("t2")], r: TestClass2
) -> TestClass3:
    return TestClass3()


def register_hooks(l: Lifecycle, t: TestClass3):
    l.append_hook(
        Hook(
            on_start=lambda: {t.run()},
        )
    )


def register_hooks_no_hint(l: Lifecycle, t):
    l.append_hook(
        Hook(
            on_start=lambda: {t.run()},
        )
    )


class TestApp(unittest.TestCase):
    def test_annotated(self):
        def test_with_annotation(
                p: Annotated[TestClass1, Annotations("t1")]) -> Annotated[TestClass3, Annotations("t2")]:
            return TestClass3()

        def register_hooks_annotated(l: Lifecycle, t: Annotated[TestClass3, Annotations("t2")]):
            l.append_hook(
                Hook(
                    on_start=lambda: {t.run()},
                )
            )
        try:
            app = App(
                Provide(
                    # test named provider
                    Provider(TestClass1, "t1"),
                    test_with_annotation,
                ),
                Invoke(
                    register_hooks_annotated,
                ),
            )
            app.run()
        except Exception as e:
            self.fail("Exception: " + str(e))
        self.assertTrue(True)

    def test_app(self):
        try:
            app = App(
                Provide(
                    # test named provider
                    Provider(TestClass1, "t1"),
                    Provider(TestClass1, "t2"),
                    TestClass2,
                    new_test_class_3,
                ),
                Invoke(
                    register_hooks,
                ),
            )
            app.run()
        except Exception as e:
            self.fail("Exception: " + str(e))
        self.assertTrue(True)

    def test_missing_dependency(self):
        exc = None
        try:
            app = App(
                Provide(),
                Invoke(
                    register_hooks,
                ),
            )

            app.run()
        except Exception as e:
            exc = e

        self.assertIsInstance(exc, MissingDependencyError)

    def test_dependency_type_error(self):
        exc = None
        try:
            provides = Provide(
                Provider(TestClass1, "t1"),
                Provider(TestClass2, "t2"),
                TestClass2,
                new_test_class_3,
            )
            app = App(
                provides,
                Invoke(
                    register_hooks,
                ),
            )

            app.run()
        except Exception as e:
            exc = e

        self.assertIsInstance(exc, DependencyTypeError)

    def test_missing_hint_error(self):
        def new_test_class_3_no_hint(
            p: Annotated[TestClass1, Annotations("t1")], p2: Annotated[TestClass1, Annotations("t2")], r: TestClass2
        ):
            return TestClass3()

        exc = None
        try:
            provides = Provide(
                Provider(TestClass1, "t1"),
                Provider(TestClass1, "t2"),
                TestClass2,
                new_test_class_3_no_hint,
            )
            app = App(
                provides,
                Invoke(
                    register_hooks,
                ),
            )

            app.run()
        except Exception as e:
            exc = e

        self.assertIsInstance(exc, MissingHintError)

    def test_missing_hint_error_input_invoke(self):
        exc = None
        try:
            provides = Provide(
                Provider(TestClass1, "t1"),
                Provider(TestClass1, "t2"),
                TestClass2,
                new_test_class_3,
            )
            app = App(
                provides,
                Invoke(
                    register_hooks_no_hint,
                ),
            )

            app.run()
        except Exception as e:
            exc = e

        self.assertIsInstance(exc, MissingHintError)

    def test_missing_hint_error_input_provide(self):
        def new_test_class_3_no_hint_input(
            p: Annotated[TestClass1, Annotations("t1")], p2: Annotated[TestClass1, Annotations("t2")], r
        ) -> TestClass3:
            return TestClass3()

        exc = None
        try:
            provides = Provide(
                Provider(TestClass1, "t1"),
                Provider(TestClass1, "t2"),
                TestClass2,
                new_test_class_3_no_hint_input,
            )
            app = App(
                provides,
                Invoke(
                    register_hooks,
                ),
            )

            app.run()
        except Exception as e:
            exc = e

        self.assertIsInstance(exc, MissingHintError)

    def test_group(self):
        def new_test_class_3_group(
            p: Annotated[TestClass1, Annotations(group=True)], r: TestClass2
        ) -> TestClass3:
            return TestClass3()

        try:
            app = App(
                Provide(
                    Provider(TestClass1, group=True),
                    Provider(TestClass1, group=True),
                    TestClass2,
                    new_test_class_3_group,
                ),
                Invoke(
                    register_hooks,
                ),
            )
            app.run()
        except Exception as e:
            self.fail("Exception: " + str(e))
        self.assertTrue(True)

    def test_modularity(self):
        m1 = Module(
            Provide(
                Provider(TestClass1, name="t1"),
                Provider(TestClass1, name="t2"),
            )
        )
        m2 = Module(
            Provide(
                TestClass2,
                new_test_class_3,
            ),
            Invoke(
                register_hooks,
            ),
        )
        try:
            app = App(Module(m1, m2))
            app.run()
        except Exception as e:
            self.fail("Exception: " + str(e))
        self.assertTrue(True)

    def test_callable_with_provider(self):
        def new_test_class_3_named(p: TestClass1) -> Annotated[TestClass3, Annotations(name="t3")]:
            return TestClass3()

        def register_hooks_named(l: Lifecycle, t: Annotated[TestClass3, Annotations(name="t3")]):
            l.append_hook(
                Hook(
                    on_start=lambda: {t.run()},
                )
            )

        try:
            app = App(
                Module(
                    Provide(
                        TestClass1,
                        new_test_class_3_named,
                    ),
                    Invoke(
                        register_hooks_named,
                    ),
                )
            )
            app.run()
        except Exception as e:
            self.fail("Exception: " + str(e))
        self.assertTrue(True)

    def test_cyclic_dependency(self):
        def new_test_class_3_cyclic(p: TestClass3) -> TestClass3:
            return TestClass3()

        exc = None
        try:
            app = App(
                Provide(new_test_class_3_cyclic),
                Invoke(
                    register_hooks,
                ),
            )

            app.run()
        except Exception as e:
            exc = e

        self.assertIsInstance(exc, RecursionError)

    def test_inherited(self):
        class TestClass4(TestClass3):
            pass

        def register_hooks(l: Lifecycle, t: TestClass3):
            l.append_hook(
                Hook(
                    on_start=lambda: {t.run()},
                )
            )

        try:
            app = App(
                Provide(
                    Provider(TestClass4, name=TestClass3),
                ),
                Invoke(
                    register_hooks,
                ),
            )

            app.run()
        except Exception as e:
            self.fail("Exception: " + str(e))
        self.assertTrue(True)


if __name__ == "__main__":
    unittest.main()
